@@ -1,11 +1,7 @@
 /* Copyright © Los Alamos National Security, LLC, and others. */
 
-/* Notes:
-
-   1. This program does not bother to free memory allocations, since they are
-      modest and the program is short-lived.
-   2. If you change any of the setuid code, consult the FAQ for some important
-      design goals. */
+/* Note: This program does not bother to free memory allocations, since they
+   are modest and the program is short-lived. */
 
 #define _GNU_SOURCE
 #include <argp.h>
@@ -14,12 +10,16 @@
 #include <fcntl.h>
 #include <libgen.h>
 #include <sched.h>
+#include <semaphore.h>
 #include <stdbool.h>
 #include <stdlib.h>
 #include <string.h>
+#include <sys/mman.h>
 #include <sys/mount.h>
+#include <sys/prctl.h>
 #include <sys/stat.h>
 #include <sys/syscall.h>
+#include <time.h>
 #include <unistd.h>
 
 #include "charliecloud.h"
@@ -47,6 +47,29 @@ const char * DEFAULT_BINDS[] = { "/dev",
 /* Log the current UIDs. */
 #define LOG_IDS log_ids(__func__, __LINE__)
 
+/* Environment variables used for --join parameters. */
+char * JOIN_CT_ENV[] =  { "OMPI_COMM_WORLD_LOCAL_SIZE",
+                          "SLURM_STEP_TASKS_PER_NODE",
+                          "SLURM_CPUS_ON_NODE",
+                          NULL };
+char * JOIN_TAG_ENV[] = { "SLURM_STEP_ID",
+                          NULL };
+
+/* Timeout in seconds for waiting for join semaphore. */
+#define JOIN_TIMEOUT 30
+
+/* Variables for coordinating --join. */
+struct {
+   bool winner_p;
+   char * sem_name;
+   sem_t * sem;
+   char * shm_name;
+   struct {
+      pid_t winner_pid;  // access anytime after initialization (write-once)
+      int proc_left_ct;  // access only while serial
+   } * shared;
+} join;
+
 
 /** Command line options **/
 
@@ -66,20 +89,17 @@ const char args_doc[] = "NEWROOT CMD [ARG...]";
 const struct argp_option options[] = {
    { "bind",        'b', "SRC[:DST]", 0,
      "mount SRC at guest DST (default /mnt/0, /mnt/1, etc.)"},
-   { "write",       'w', 0,     0, "mount image read-write"},
-   { "no-home",      -2, 0,     0, "do not bind-mount your home directory"},
    { "cd",          'c', "DIR", 0, "initial working directory in container"},
-#ifndef SETUID
    { "gid",         'g', "GID", 0, "run as GID within container" },
-#endif
-   { "is-setuid",    -1, 0,     0,
-     "exit successfully if compiled for setuid, else fail" },
+   { "join",        'j', 0,     0, "use same container as peer ch-run" },
+   { "join-ct",      -3, "N",   0, "number of ch-run peers (implies --join)" },
+   { "join-tag",     -4, "TAG", 0, "label for peer group (implies --join)" },
+   { "no-home",      -2, 0,     0, "do not bind-mount your home directory"},
    { "private-tmp", 't', 0,     0, "use container-private /tmp" },
-#ifndef SETUID
    { "uid",         'u', "UID", 0, "run as UID within container" },
-#endif
    { "verbose",     'v', 0,     0, "be more verbose (debug if repeated)" },
    { "version",     'V', 0,     0, "print version and exit" },
+   { "write",       'w', 0,     0, "mount image read-write"},
    { 0 }
 };
 
@@ -93,26 +113,32 @@ struct args {
    gid_t container_gid;
    uid_t container_uid;
    char * newroot;
-   char * initial_working_dir;
+   char * initial_dir;
+   bool join;
+   int join_ct;
+   char * join_tag;
    bool private_home;
    bool private_tmp;
    bool writable;
    int user_cmd_start;  // index into argv where NEWROOT is
-   int verbose;
 };
 
 void enter_udss(char * newroot, bool writeable, struct bind * binds,
                 bool private_tmp, bool private_home);
+bool get_first_env(char ** array, char ** name, char ** value);
+void join_begin();
+int join_ct();
+void join_end();
+void join_namespace(pid_t pid, char * ns);
+void join_namespaces();
+char * join_tag();
 void log_ids(const char * func, int line);
 void run_user_command(int argc, char * argv[], int user_cmd_start);
 static error_t parse_opt(int key, char * arg, struct argp_state * state);
+int parse_int(char * s, bool extra_ok, char * error_tag);
 void privs_verify_invoking();
+void sem_timedwait_relative(sem_t * sem, int timeout);
 void setup_namespaces(uid_t cuid, gid_t cgid);
-#ifdef SETUID
-void privs_drop_permanently();
-void privs_drop_temporarily();
-void privs_restore();
-#endif
 
 struct args args;
 const struct argp argp = { options, parse_opt, args_doc, usage };
@@ -123,37 +149,46 @@ const struct argp argp = { options, parse_opt, args_doc, usage };
 int main(int argc, char * argv[])
 {
    privs_verify_invoking();
-#ifdef SETUID
-   privs_drop_temporarily();
-#endif
    memset(args.binds, 0, sizeof(args.binds));
-   args.container_gid = getgid();
-   args.container_uid = getuid();
-   args.initial_working_dir = NULL;
+   args.container_gid = getegid();
+   args.container_uid = geteuid();
+   args.initial_dir = NULL;
+   args.join = false;
+   args.join_ct = 0;
+   args.join_tag = NULL;
    args.private_home = false;
    args.private_tmp = false;
-   args.verbose = 0;
+   verbose = 1;  // in charliecloud.c
    Z_ (setenv("ARGP_HELP_FMT", "opt-doc-col=25,no-dup-args-note", 0));
    Z_ (argp_parse(&argp, argc, argv, 0, &(args.user_cmd_start), &args));
    Te (args.user_cmd_start < argc - 1, "NEWROOT and/or CMD not specified");
    assert(args.binds[USER_BINDS_MAX].src == NULL);  // overrun in argp_parse?
-   args.newroot = realpath(argv[args.user_cmd_start++], NULL);
-   Tf (args.newroot != NULL, "couldn't resolve image path");
-
-   if (args.verbose) {
-      fprintf(stderr, "newroot: %s\n", args.newroot);
-      fprintf(stderr, "container uid: %u\n", args.container_uid);
-      fprintf(stderr, "container gid: %u\n", args.container_gid);
-      fprintf(stderr, "private /tmp: %d\n", args.private_tmp);
+   args.newroot = realpath(argv[args.user_cmd_start], NULL);
+   Tf (args.newroot != NULL, "can't find image: %s", argv[args.user_cmd_start]);
+   args.user_cmd_start++;
+   if (args.join) {
+      args.join_ct = join_ct();
+      args.join_tag = join_tag();
    }
 
-   setup_namespaces(args.container_uid, args.container_gid);
-   enter_udss(args.newroot, args.writable, args.binds,
-              args.private_tmp, args.private_home);
+   INFO("verbosity: %d", verbose);
+   INFO("newroot: %s", args.newroot);
+   INFO("container uid: %u", args.container_uid);
+   INFO("container gid: %u", args.container_gid);
+   INFO("join: %d %d %s", args.join, args.join_ct, args.join_tag);
+   INFO("private /tmp: %d", args.private_tmp);
+
+   if (args.join)
+      join_begin();
+   if (!args.join || join.winner_p) {
+      setup_namespaces(args.container_uid, args.container_gid);
+      enter_udss(args.newroot, args.writable, args.binds,
+                 args.private_tmp, args.private_home);
+   } else
+      join_namespaces();
+   if (args.join)
+      join_end();
 
-#ifdef SETUID
-   privs_drop_permanently();
-#endif
    run_user_command(argc, argv, args.user_cmd_start); // should never return
    exit(EXIT_FAILURE);
 }
@@ -178,20 +213,6 @@ void enter_udss(char * newroot, bool writable, struct bind * binds,
 
    LOG_IDS;
 
-#ifdef SETUID
-
-   privs_restore();
-
-   // Make the whole filesystem tree private. Otherwise, there's a big mess,
-   // as the manipulations of the shared mounts propagate into the parent
-   // namespace. Then the mount(MS_MOVE) call below fails with EINVAL, and
-   // nothing is cleaned up so the mounts are a big tangle and ch-tar2dir will
-   // delete your home directory. I think this is redundant with some of the
-   // below, but it doesn't seem to hurt.
-   Z_ (mount(NULL, "/", NULL, MS_REC | MS_PRIVATE, NULL));
-
-#endif
-
    // Claim newroot for this namespace
    Zf (mount(newroot, newroot, NULL,
              MS_REC | MS_BIND | MS_PRIVATE, NULL), newroot);
@@ -265,14 +286,167 @@ void enter_udss(char * newroot, bool writable, struct bind * binds,
    Zf (syscall(SYS_pivot_root, newroot, path), "can't pivot_root(2)");
    Zf (chroot("."), "can't chroot(2) into new root");
    Zf (umount2("/dev", MNT_DETACH), "can't umount old root");
+}
 
-#ifdef SETUID
-   privs_drop_temporarily();
-#endif
+/* Find the first environment variable in array that is set; put its name in
+   *name and its value in *value, and return true. If none are set, return
+   false, and *name and *value are undefined. */
+bool get_first_env(char ** array, char ** name, char ** value)
+{
+   for (int i = 0; array[i] != NULL; i++) {
+      *name = array[i];
+      *value = getenv(*name);
+      if (*value != NULL)
+         return true;
+   }
 
-   if (args.initial_working_dir != NULL)
-      Zf (chdir(args.initial_working_dir),
-          "can't cd to %s", args.initial_working_dir);
+   return false;
+}
+
+/* Begin coordinated section of namespace joining. */
+void join_begin()
+{
+   int fd;
+
+   T_ (1 <= asprintf(&join.sem_name, "/ch-run_%s", args.join_tag));
+   T_ (1 <= asprintf(&join.shm_name, "/ch-run_%s", args.join_tag));
+
+   // Serialize.
+   join.sem = sem_open(join.sem_name, O_CREAT, 0600, 1);
+   T_ (join.sem != SEM_FAILED);
+   sem_timedwait_relative(join.sem, JOIN_TIMEOUT);
+
+   // Am I the winner?
+   fd = shm_open(join.shm_name, O_CREAT|O_EXCL|O_RDWR, 0600);
+   if (fd > 0) {
+      INFO("join: I won");
+      join.winner_p = true;
+      Z_ (ftruncate(fd, sizeof(*join.shared)));
+   } else if (errno == EEXIST) {
+      join.winner_p = false;
+      fd = shm_open(join.shm_name, O_RDWR, 0);
+      T_ (fd > 0);
+   } else {
+      T_ (0);
+   }
+
+   join.shared = mmap(NULL, sizeof(*join.shared), PROT_READ|PROT_WRITE,
+                      MAP_SHARED, fd, 0);
+   T_ (join.shared != NULL);
+   Z_ (close(fd));
+
+   if (join.winner_p) {
+      join.shared->winner_pid = getpid();
+      join.shared->proc_left_ct = args.join_ct;
+      // Keep lock; winner still serialized.
+   } else {
+      INFO("join: winner pid: %d", join.shared->winner_pid);
+      Z_ (sem_post(join.sem));
+      // Losers run in parallel (winner will be done by now).
+   }
+}
+
+/* Find an appropriate join count; assumes --join was specified or implied.
+   Exit with error if no valid value is available. */
+int join_ct()
+{
+   int j = 0;
+   char * ev_name, * ev_value;
+
+   if (args.join_ct != 0) {
+      INFO("join: peer group size from command line");
+      j = args.join_ct;
+      goto end;
+   }
+
+   if (get_first_env(JOIN_CT_ENV, &ev_name, &ev_value)) {
+      INFO("join: peer group size from %s", ev_name);
+      j = parse_int(ev_value, true, ev_name);
+      goto end;
+   }
+
+end:
+   Te(j > 0, "join: no valid peer group size found");
+   return j;
+}
+
+/* End coordinated section of namespace joining. */
+void join_end()
+{
+   // Serialize (winner never released lock).
+   if (!join.winner_p)
+      sem_timedwait_relative(join.sem, JOIN_TIMEOUT);
+
+   join.shared->proc_left_ct--;
+   INFO("join: %d peers left excluding myself", join.shared->proc_left_ct);
+
+   if (join.shared->proc_left_ct <= 0) {
+      INFO("join: cleaning up IPC resources");
+      Te (join.shared->proc_left_ct == 0, "expected 0 peers left but found %d",
+          join.shared->proc_left_ct);
+      Zf (sem_unlink(join.sem_name), "can't unlink sem: %s", join.sem_name);
+      Zf (shm_unlink(join.shm_name), "can't unlink shm: %s", join.shm_name);
+   }
+
+   // Parallelize.
+   Z_ (sem_post(join.sem));
+
+   Z_ (munmap(join.shared, sizeof(*join.shared)));
+   Z_ (sem_close(join.sem));
+
+   INFO("join: done");
+}
+
+/* Join a specific namespace. */
+void join_namespace(pid_t pid, char * ns)
+{
+   char * path;
+   int fd;
+
+   T_ (1 <= asprintf(&path, "/proc/%d/ns/%s", pid, ns));
+   fd = open(path, O_RDONLY);
+   if (fd == -1) {
+      if (errno == ENOENT) {
+         Te (0, "join: %s not found; is winner still running?", path);
+      } else {
+         Tf (0, "join: can't open %s", path);
+      }
+   }
+   Zf (setns(fd, 0), "can't join %s namespace of pid %d", ns, pid);
+}
+
+/* Join the existing namespaces created by the join winner. */
+void join_namespaces()
+{
+   join_namespace(join.shared->winner_pid, "user");
+   join_namespace(join.shared->winner_pid, "mnt");
+}
+
+/* Find an appropriate join tag; assumes --join was specified or implied. Exit
+   with error if no valid value is found. */
+char * join_tag()
+{
+   char * tag;
+   char * ev_name, * ev_value;
+
+   if (args.join_tag != NULL) {
+      INFO("join: peer group tag from command line");
+      tag = args.join_tag;
+      goto end;
+   }
+
+   if (get_first_env(JOIN_TAG_ENV, &ev_name, &ev_value)) {
+      INFO("join: peer group tag from %s", ev_name);
+      tag = ev_value;
+      goto end;
+   }
+
+   INFO("join: peer group tag from getppid(2)");
+   T_ (1 <= asprintf(&tag, "%d", getppid()));
+
+end:
+   Te(tag[0] != '\0', "join: peer group tag cannot be empty string");
+   return tag;
 }
 
 /* If verbose, print uids and gids on stderr prefixed with where. */
@@ -283,7 +457,7 @@ void log_ids(const char * func, int line)
    gid_t supp_gids[SUPP_GIDS_MAX];
    int supp_gid_ct;
 
-   if (args.verbose >= 2) {
+   if (verbose >= 3) {
       Z_ (getresuid(&ruid, &euid, &suid));
       Z_ (getresgid(&rgid, &egid, &sgid));
       fprintf(stderr, "%s %d: uids=%d,%d,%d, gids=%d,%d,%d + ", func, line,
@@ -302,26 +476,44 @@ void log_ids(const char * func, int line)
    }
 }
 
+/* Parse an integer string arg and return the result. If an error occurs,
+   print a message prefixed by error_tag and exit. If not extra_ok, additional
+   characters remaining after the integer are an error. */
+int parse_int(char * s, bool extra_ok, char * error_tag)
+{
+   char * end;
+   long l;
+
+   errno = 0;
+   l = strtol(s, &end, 10);
+   Tf (errno == 0, error_tag);
+   Ze (end == s, "%s: no digits found", error_tag);
+   if (!extra_ok)
+      Te (*end == 0, "%s: extra characters after digits", error_tag);
+   Te (l >= INT_MIN && l <= INT_MAX, "%s: out of range", error_tag);
+   return (int)l;
+}
+
 /* Parse one command line option. Called by argp_parse(). */
 static error_t parse_opt(int key, char * arg, struct argp_state * state)
 {
    struct args * as = state->input;
    int i;
-   long l;
 
    switch (key) {
-   case -1:
-#ifdef SETUID
-      exit(EXIT_SUCCESS);
-#else
-      exit(EXIT_FAILURE);
-#endif
-      break;
-   case -2:
+   case -2: // --private-home
       as->private_home = true;
       break;
+   case -3: // --join-ct
+      as->join = true;
+      as->join_ct = parse_int(arg, false, "--join-ct");
+      break;
+   case -4: // --join-tag
+      as->join = true;
+      as->join_tag = arg;
+      break;
    case 'c':
-      as->initial_working_dir = arg;
+      as->initial_dir = arg;
       break;
    case 'b':
       for (i = 0; as->binds[i].src != NULL; i++)
@@ -338,26 +530,28 @@ static error_t parse_opt(int key, char * arg, struct argp_state * state)
       Te (as->binds[i].dst[0] != 0, "--bind: no destination provided");
       break;
    case 'g':
-      errno = 0;
-      l = strtol(arg, NULL, 0);
-      Te (errno == 0 && l >= 0, "GID must be a non-negative integer");
-      as->container_gid = (gid_t)l;
+      i = parse_int(arg, false, "--gid");
+      Te (i >= 0, "--gid: must be non-negative");
+      as->container_gid = (gid_t) i;
+      break;
+   case 'j':
+      as->join = true;
       break;
    case 't':
       as->private_tmp = true;
       break;
    case 'u':
-      errno = 0;
-      l = strtol(arg, NULL, 0);
-      Te (errno == 0 && l >= 0, "UID must be a non-negative integer");
-      as->container_uid = (uid_t)l;
+      i = parse_int(arg, false, "--uid");
+      Te (i >= 0, "--uid: must be non-negative");
+      as->container_uid = (uid_t) i;
       break;
    case 'V':
       version();
       exit(EXIT_SUCCESS);
       break;
    case 'v':
-      as->verbose++;
+      verbose++;
+      Te(verbose <= 3, "--verbose can be specified at most twice");
       break;
    case 'w':
       as->writable = true;
@@ -392,70 +586,11 @@ void privs_verify_invoking()
    T_ (egid != 0);                           // no privilege
    T_ (egid == rgid && egid == sgid);        // no setuid or funny business
 
-   // Setuid must match the compiled mode.
-#ifdef SETUID
-   T_ (ruid != 0 && euid == 0 && suid == 0); // must be setuid root
-#else
+   // No UID privilege allowed either.
    T_ (euid != 0);                           // no privilege
    T_ (euid == ruid && euid == suid);        // no setuid or funny business
-#endif
 }
 
-/* Drop UID privileges permanently. */
-#ifdef SETUID
-void privs_drop_permanently()
-{
-   uid_t uid_wanted, ruid, euid, suid;
-   gid_t rgid, egid, sgid;
-
-   // Drop privileges.
-   uid_wanted = getuid();
-   T_ (uid_wanted != 0);  // abort if real UID is root
-   Z_ (setresuid(uid_wanted, uid_wanted, uid_wanted));
-
-   // Try to regain privileges; it should fail.
-   T_ (-1 == setuid(0));
-   T_ (-1 == setresuid(-1, 0, -1));
-
-   // UIDs should be unprivileged and the same.
-   Z_ (getresuid(&ruid, &euid, &suid));
-   T_ (ruid == uid_wanted);
-   T_ (uid_wanted == ruid && uid_wanted == euid && uid_wanted == suid);
-
-   // GIDs should be unprivileged and the same.
-   Z_ (getresgid(&rgid, &egid, &sgid));
-   T_ (rgid != 0);
-   T_ (rgid == egid && rgid == sgid);
-}
-#endif // SETUID
-
-/* Drop UID privileges temporarily; can be regained with privs_restore(). */
-#ifdef SETUID
-void privs_drop_temporarily()
-{
-   uid_t unpriv_uid = getuid();
-
-   if (unpriv_uid == 0) {
-      // Invoked as root, so descend to nobody.
-      unpriv_uid = 65534;
-   }
-
-   Z_ (setresuid(-1, unpriv_uid, -1));
-   T_ (unpriv_uid == geteuid());
-}
-#endif // SETUID
-
-/* Restore privileges that have been dropped with privs_drop_temporarily(). */
-#ifdef SETUID
-void privs_restore()
-{
-   uid_t ruid, euid, suid;
-
-   Z_ (setresuid(-1, 0, -1));
-   Z_ (getresuid(&ruid, &euid, &suid));
-}
-#endif // SETUID
-
 /* Replace the current process with user command and arguments. argv will be
    overwritten in order to avoid the need for copying it, because execvp()
    requires null-termination instead of an argument count. */
@@ -469,45 +604,52 @@ void run_user_command(int argc, char * argv[], int user_cmd_start)
       argv[i - user_cmd_start] = argv[i];
    argv[argc - user_cmd_start] = NULL;
 
+   // --cd
+   if (args.initial_dir != NULL)
+      Zf (chdir(args.initial_dir), "can't cd to %s", args.initial_dir);
+
    // Append /bin to $PATH if not already present. See FAQ.
    old_path = getenv("PATH");
    if (old_path == NULL) {
-      if (args.verbose)
-         fprintf(stderr, "warning: $PATH not set\n");
+      WARNING("$PATH not set");
    } else if (   strstr(old_path, "/bin") != old_path
               && !strstr(old_path, ":/bin")) {
       T_ (1 <= asprintf(&new_path, "%s:/bin", old_path));
       Z_ (setenv("PATH", new_path, 1));
-      if (args.verbose)
-         fprintf(stderr, "new $PATH: %s\n", new_path);
+      INFO("new $PATH: %s", new_path);
    }
 
-   if (args.verbose) {
+   if (verbose >= 3) {
       fprintf(stderr, "cmd at %d/%d:", user_cmd_start, argc);
       for (int i = 0; argv[i] != NULL; i++)
          fprintf(stderr, " %s", argv[i]);
       fprintf(stderr, "\n");
    }
 
+   Zf (prctl(PR_SET_NO_NEW_PRIVS, 1, 0, 0, 0), "can't set no_new_privs");
    execvp(argv[0], argv);  // only returns if error
-   Tf (0, "can't execve(2) user command");
+   Tf (0, "can't execve(2): %s", argv[0]);
+}
+
+/* Wait for semaphore sem for up to timeout seconds. If timeout or an error,
+   exit unsuccessfully. */
+void sem_timedwait_relative(sem_t * sem, int timeout)
+{
+   struct timespec deadline;
+
+   // sem_timedwait() requires a deadline rather than a timeout.
+   Z_ (clock_gettime(CLOCK_REALTIME, &deadline));
+   deadline.tv_sec += timeout;
+
+   if (sem_timedwait(sem, &deadline)) {
+      Ze (errno == ETIMEDOUT, "timeout waiting for join lock");
+      Tf (0, "failure waiting for join lock");
+   }
 }
 
 /* Activate the desired isolation namespaces. */
 void setup_namespaces(uid_t cuid, gid_t cgid)
 {
-#ifdef SETUID
-
-   // can't change IDs from invoking
-   T_ (cuid == getuid());
-   T_ (cgid == getgid());
-
-   privs_restore();
-   Z_ (unshare(CLONE_NEWNS));
-   privs_drop_temporarily();
-
-#else // not SETUID
-
    int fd;
    uid_t euid = -1;
    gid_t egid = -1;
@@ -516,7 +658,7 @@ void setup_namespaces(uid_t cuid, gid_t cgid)
    egid = getegid();
 
    LOG_IDS;
-   Z_ (unshare(CLONE_NEWNS|CLONE_NEWUSER));
+   Zf (unshare(CLONE_NEWNS|CLONE_NEWUSER), "can't init user+mount namespaces");
    LOG_IDS;
 
    /* Write UID map. What we are allowed to put here is quite limited. Because
@@ -541,6 +683,4 @@ void setup_namespaces(uid_t cuid, gid_t cgid)
    T_ (1 <= dprintf(fd, "%d %d 1\n", cgid, egid));
    Z_ (close(fd));
    LOG_IDS;
-
-#endif // not SETUID
 }
